@@ -1,7 +1,6 @@
 #Appears to be common/best practice to import required packages in every file
 #they are used in
 import numpy as np
-from scipy.ndimage import filters, morphology #For minimum finding
 
 #For ND interpolation
 # from scipy.interpolate import interpnd, RectBivariateSpline
@@ -476,7 +475,6 @@ class MinimumEnergyPath:
             
         return netForce
     
-#TODO: do we compute the action for all points after optimizing? or let someone else do that? 
 class VerletMinimization:
     """
     :Maintainer: Daniel
@@ -564,6 +562,7 @@ class VerletMinimization:
          
         vProj = np.zeros((self.nPts,self.nDims))
         
+        t0 = time.time()
         try:
             for step in range(1,maxIters+1):
                 self.allForces[step] = self.nebObj.compute_force(self.allPts[step])
@@ -585,12 +584,14 @@ class VerletMinimization:
                 self.allPts[step+1] = self.allPts[step] + self.allVelocities[step]*tStep + \
                     0.5*accel*tStep**2
         finally:
+            t1 = time.time()
             self.nebObj.logger.flush()
+            self.nebObj.logger.write_runtime(t1-t0)
             
         return None
     
     def fire(self,tStep,maxIters,fireParams={},useLocal=True,earlyStop=True,
-             earlyStopParams={}):
+             earlyStopParams={},earlyAbort=False,earlyAbortParams={}):
         """
         Wrapper for fast inertial relaxation engine.
         FIRE step taken from http://dx.doi.org/10.1103/PhysRevLett.97.170201
@@ -645,6 +646,17 @@ class VerletMinimization:
         for key in defaultStopParams.keys():
             if key not in earlyStopParams.keys():
                 earlyStopParams[key] = defaultStopParams[key]
+                
+        defaultAbortParams = {"startCheckIter":20,"nStabIters":10,"checkFreq":10,\
+                              "variance":1}
+            
+        for key in earlyAbortParams.keys():
+            if key not in defaultAbortParams.keys():
+                raise ValueError("Key "+key+" in earlyStopParams not allowed")
+                
+        for key in defaultAbortParams.keys():
+            if key not in earlyAbortParams.keys():
+                earlyAbortParams[key] = defaultAbortParams[key]
         
         self.allPts = np.zeros((maxIters+2,self.nPts,self.nDims))
         self.allVelocities = np.zeros((maxIters+1,self.nPts,self.nDims))
@@ -666,6 +678,8 @@ class VerletMinimization:
         tStepArr[0] = tStep
         alphaArr[0] = fireParams["aStart"]
         
+        endsWithoutError = True
+        
         t0 = time.time()
         try:
             for step in range(1,maxIters+1):
@@ -690,6 +704,12 @@ class VerletMinimization:
                         alphaArr = alphaArr[:step]
                         break
                     
+                if earlyAbort:
+                    stopBool = self._check_early_abort(step,earlyAbortParams)
+                    if stopBool:
+                        endsWithoutError = False
+                        break
+                    
             #Final iteration
             if useLocal:
                 tStepFinal = tStepArr[-1].reshape((-1,1))
@@ -714,7 +734,7 @@ class VerletMinimization:
             if earlyStop:
                 self.nebObj.logger.write_early_stop_params(earlyStopParams)
         
-        return tStepArr, alphaArr, stepsSinceReset
+        return tStepArr, alphaArr, stepsSinceReset, endsWithoutError
     
     def _local_fire_iter(self,step,tStepArr,alphaArr,stepsSinceReset,fireParams):
         tStepPrev = tStepArr[step-1].reshape((-1,1)) #For multiplication below
@@ -761,40 +781,84 @@ class VerletMinimization:
         
         return tStepArr, alphaArr, stepsSinceReset
 
-
     def _global_fire_iter(self,step,tStepArr,alphaArr,stepsSinceReset,fireParams):
-        self.allPts[step] = self.allPts[step-1] + \
-            tStepArr[step-1]*self.allVelocities[step-1] + \
-            0.5*self.allForces[step-1]*tStepArr[step-1]**2
-        self.allForces[step] = self.nebObj.compute_force(self.allPts[step])
+        """
+        Implements the FIRE algorithm from doi.org/10.1016/j.commatsci.2020.109584
+        (algorithm 1) using a mixed semi-implicit Euler update (algorithm 4). Note
+        that there is a typo in the mixing: line 2 should involve v(t + \Delta t) on
+        the right-hand side, rather than v(t). Line 10 of algorithm 1 is ignored;
+        performance appears to be way better (or even correct at all) when mixing
+        occurs in the integration step
+
+        Parameters
+        ----------
+        step : TYPE
+            DESCRIPTION.
+        tStepArr : TYPE
+            DESCRIPTION.
+        alphaArr : TYPE
+            DESCRIPTION.
+        stepsSinceReset : TYPE
+            DESCRIPTION.
+        fireParams : TYPE
+            DESCRIPTION.
+
+        Returns
+        -------
+        tStepArr : TYPE
+            DESCRIPTION.
+        alphaArr : TYPE
+            DESCRIPTION.
+        stepsSinceReset : TYPE
+            DESCRIPTION.
+
+        """
+        vdotf = 0.
+        for ptIter in range(self.nPts):
+            vdotf += np.dot(self.allVelocities[step-1,ptIter],self.allForces[step-1,ptIter])
         
-        #Doesn't seem to make a difference
-        #What the Wikipedia article on velocity Verlet uses
-        self.allVelocities[step] = \
-            0.5*tStepArr[step-1]*(self.allForces[step]+self.allForces[step-1])
-        #What Eric uses
-        # self.allVelocities[step] = tStepArr[step-1]*self.allForces[step]
+        if vdotf > 0:
+            stepsSinceReset += 1
+            
+            if stepsSinceReset > fireParams["nAccel"]:
+                tStepArr[step] = min(tStepArr[step-1]*fireParams["fInc"],fireParams["dtMax"])
+                alphaArr[step] = alphaArr[step-1]**fireParams["fAlpha"]
+            else:
+                tStepArr[step] = tStepArr[step-1]
+                alphaArr[step] = alphaArr[step-1]            
+        else:
+            stepsSinceReset = 0
+            self.allVelocities[step-1,ptIter] = np.zeros(self.nDims)
+            tStepArr[step] = max(tStepArr[step-1]*fireParams["fDecel"],fireParams["dtMin"])
+            alphaArr[step] = fireParams["aStart"]
+            
+        #Semi-implicit Euler integration
+        self.allVelocities[step] = self.allVelocities[step-1] + tStepArr[step]*self.allForces[step-1]
+        
+        vdotv = 0.
+        fdotf = 0.
+        for ptIter in range(self.nPts):
+            vdotv += np.linalg.norm(self.allVelocities[step,ptIter])
+            fdotf += np.linalg.norm(self.allForces[step-1,ptIter])
+        
+        if fdotf > 10**(-16):
+            scale = vdotv/fdotf
+        else:
+            scale = 0.
+        
+        self.allVelocities[step] = (1-alphaArr[step]) * self.allVelocities[step]\
+            + alphaArr[step] * scale * self.allForces[step-1]
+        
+        shift = tStepArr[step]*self.allVelocities[step]
         
         for ptIter in range(self.nPts):
-            alpha = alphaArr[step-1]
-            
-            product = np.dot(self.allVelocities[step-1,ptIter],self.allForces[step,ptIter])
-            if product > 0:
-                vMag = np.linalg.norm(self.allVelocities[step-1,ptIter])
-                fHat = self.allForces[step,ptIter]/np.linalg.norm(self.allForces[step,ptIter])
-                vp = (1-alpha)*self.allVelocities[step-1,ptIter] + alpha*vMag*fHat
-                self.allVelocities[step,ptIter] += vp
-                
-                if stepsSinceReset > fireParams["nAccel"]:
-                    tStepArr[step] = min(tStepArr[step-1]*fireParams["fInc"],fireParams["dtMax"])
-                    alphaArr[step] = alpha*fireParams["fAlpha"]
-                
-                stepsSinceReset += 1
-            else:
-                tStepArr[step] = max(tStepArr[step-1]*fireParams["fDecel"],fireParams["dtMin"])
-                # self.allVelocities[step,ptIter] = np.zeros(self.nDims)
-                alphaArr[step] = fireParams["aStart"]
-                stepsSinceReset = 0
+            for dimIter in range(self.nDims):
+                if(abs(shift[ptIter,dimIter])>fireParams["maxmove"][dimIter]):
+                    shift[ptIter] = shift[ptIter] * \
+                        fireParams["maxmove"][dimIter]/abs(shift[ptIter,dimIter])
+        
+        self.allPts[step] = self.allPts[step-1] + shift
+        self.allForces[step] = self.nebObj.compute_force(self.allPts[step])
         
         return tStepArr, alphaArr, stepsSinceReset
     
@@ -806,10 +870,6 @@ class VerletMinimization:
         
         Velocity update taken from 
         https://en.wikipedia.org/wiki/Verlet_integration#Velocity_Verlet
-        
-        TODO: consider making FIRE its own class, or allowing for attributes
-        like fireParams and etc
-        TODO: add maxmove parameter to prevent path exploding
 
         Parameters
         ----------
@@ -833,8 +893,9 @@ class VerletMinimization:
         """
         
         defaultFireParams = \
-            {"dtMax":10.,"dtMin":0.001,"nAccel":10,"fInc":1.1,"fAlpha":0.99,\
-             "fDecel":0.5,"aStart":0.1,"maxmove":np.full(self.nDims,1.0)}
+            {"dtMax":10.,"dtMin":0.02,"nAccel":10,"fInc":1.1,"fAlpha":0.99,\
+             "fDecel":0.5,"aStart":0.1,"maxmove":np.full(self.nDims,1.0),\
+             "minDecelIter":20}
             
         for key in fireParams.keys():
             if key not in defaultFireParams.keys():
@@ -863,8 +924,7 @@ class VerletMinimization:
         self.allForces[0] = self.nebObj.compute_force(self.allPts[0])
         
         if useLocal:
-            tStepArr = np.zeros((maxIters+1,self.nPts))
-            tStepArr[:,:] = tStep
+            tStepArr = tStep*np.ones((maxIters+1,self.nPts))
             alphaArr = np.zeros((maxIters+1,self.nPts))
             stepsSinceReset = np.zeros((self.nPts))
         else:
@@ -878,15 +938,14 @@ class VerletMinimization:
         t0 = time.time()
         try:
             for step in range(1,maxIters+1):
-                #TODO: check potential off-by-one indexing on tStep
                 if useLocal:
                     tStepArr,alphaArr,stepsSinceReset = \
                         self._local_fire2_iter(step,tStepArr,alphaArr,stepsSinceReset,\
-                                              fireParams)
+                                               fireParams)
                 else:
                     tStepArr,alphaArr,stepsSinceReset = \
                         self._global_fire2_iter(step,tStepArr,alphaArr,stepsSinceReset,\
-                                               fireParams)
+                                                fireParams)
                             
                 if earlyStop:
                     stopBool = self._check_early_stop(step,earlyStopParams)
@@ -925,104 +984,114 @@ class VerletMinimization:
         
         return tStepArr, alphaArr, stepsSinceReset
     
+    def _local_fire2_iter(self,step,tStepArr,alphaArr,stepsSinceReset,fireParams):
+        warnings.warn("Local FIRE2 currently calls local FIRE update")
+        return self._local_fire_iter(step,tStepArr,alphaArr,stepsSinceReset,fireParams)
+    
     def _global_fire2_iter(self,step,tStepArr,alphaArr,stepsSinceReset,fireParams):
-        self.allForces[0] = self.nebObj.compute_force(self.allPts[step-1])
+        """
+        Implements the FIRE 2 algorithm from doi.org/10.1016/j.commatsci.2020.109584
+        (algorithm 2) using a mixed semi-implicit Euler update (algorithm 4). Note
+        that there is a typo in the mixing: line 2 should involve v(t + \Delta t) on
+        the right-hand side, rather than v(t)
+
+        Parameters
+        ----------
+        step : TYPE
+            DESCRIPTION.
+        tStepArr : TYPE
+            DESCRIPTION.
+        alphaArr : TYPE
+            DESCRIPTION.
+        stepsSinceReset : TYPE
+            DESCRIPTION.
+        fireParams : TYPE
+            DESCRIPTION.
+
+        Returns
+        -------
+        tStepArr : TYPE
+            DESCRIPTION.
+        alphaArr : TYPE
+            DESCRIPTION.
+        stepsSinceReset : TYPE
+            DESCRIPTION.
+
+        """
         vdotf = 0.0
         for ptIter in range(self.nPts):
-            vdotf += np.dot(self.allVelocities[0,ptIter],self.allForces[0,ptIter])
+            vdotf += np.dot(self.allVelocities[step-1,ptIter],self.allForces[step-1,ptIter])
 
-        if(vdotf > 0.0):
-            vdotv = 0.0
-            fdotf = 0.0
-            for ptIter in range(self.nPts):
-                vdotv += np.dot(self.allVelocities[0,ptIter],self.allVelocities[0,ptIter])
-                fdotf += np.dot(self.allForces[0,ptIter],self.allForces[0,ptIter])
-            scale = 1.0 - alphaArr[0]
-            scale2= alphaArr[0]*np.sqrt(vdotv/fdotf)
-            if(fdotf <= 2e-20): scale2 = 0.0
-            if stepsSinceReset > fireParams["nAccel"]:
-                tStepArr[0] = \
-                    min(tStepArr[0]*fireParams["fInc"],fireParams["dtMax"])
-                alphaArr[0] = alphaArr[0]*fireParams["fAlpha"]
-            else:
-                tStepArr[0] = tStepArr[0]
-                
+        if vdotf > 0.0:
             stepsSinceReset += 1
-        else:
-            if(step > 20):
-                alphaArr[0] = fireParams["aStart"]
-                tStepArr[0] = \
-                    max(tStepArr[0]*fireParams["fDecel"],fireParams["dtMin"])
-            self.allPts[step,:,:] = self.allPts[step-1] - 0.5*tStepArr[0]*self.allVelocities[0,:,:]
-            self.allVelocities[0,:,:] = 0.0
-            #v0 = True
-
-        #if (v0):
-        #    self.allVelocities[0,:,:] = tStepArr[0]*self.allForces[0,:,:]
-
-        self.allVelocities[0,:,:] += tStepArr[0]*self.allForces[0,:,:]
-        if(vdotf > 0.0):
-            self.allVelocities[0,:,:] = scale*self.allVelocities[0,:,:] + \
-                scale2*self.allForces[0,:,:]
-        shift = tStepArr[0]*self.allVelocities[0,:,:]
-
-        for ptIter in range(self.nPts):
-            for dimIter in range(self.nDims):
-                if(abs(shift[ptIter,dimIter])>fireParams["maxmove"][dimIter]):
-                    shift[ptIter] = shift[ptIter] * \
-                        fireParams["maxmove"][dimIter]/abs(shift[ptIter,dimIter])
-
-        self.allPts[step] = self.allPts[step-1] + shift
-
-        return tStepArr, alphaArr, stepsSinceReset
-
-
-    def _local_fire2_iter(self,step,tStepArr,alphaArr,stepsSinceReset,fireParams):
-        tStepPrev = tStepArr[step-1].reshape((-1,1)) #For multiplication below
-        
-        shift = tStepPrev*self.allVelocities[step-1] + \
-                0.5*self.allForces[step-1]*tStepPrev**2
-
-        for ptIter in range(self.nPts):
-            for dimIter in range(self.nDims):
-                if(abs(shift[ptIter,dimIter])>fireParams["maxmove"][dimIter]):
-                    shift[ptIter] = shift[ptIter] * \
-                        fireParams["maxmove"][dimIter]/abs(shift[ptIter,dimIter])
-
-        self.allPts[step] = self.allPts[step-1] + shift
-        
-        self.allForces[step] = self.nebObj.compute_force(self.allPts[step])
-        #What the Wikipedia article on velocity Verlet uses
-        self.allVelocities[step] = \
-            0.5*tStepPrev*(self.allForces[step]+self.allForces[step-1])
-        
-        for ptIter in range(self.nPts):
-            alpha = alphaArr[step-1,ptIter]
-            
-            product = np.dot(self.allVelocities[step-1,ptIter],self.allForces[step,ptIter])
-            if product > 0:
-                vMag = np.linalg.norm(self.allVelocities[step-1,ptIter])
-                fHat = self.allForces[step,ptIter]/np.linalg.norm(self.allForces[step,ptIter])
-                self.allVelocities[step,ptIter] += (1-alpha)*self.allVelocities[step-1,ptIter] + \
-                    alpha*vMag*fHat
-                
-                if stepsSinceReset[ptIter] > fireParams["nAccel"]:
-                    tStepArr[step,ptIter] = \
-                        min(tStepArr[step-1,ptIter]*fireParams["fInc"],fireParams["dtMax"])
-                    alphaArr[step,ptIter] = alpha*fireParams["fAlpha"]
-                else:
-                    tStepArr[step,ptIter] = tStepArr[step-1,ptIter]
-                
-                stepsSinceReset[ptIter] += 1
+            if stepsSinceReset > fireParams["nAccel"]:
+                tStepArr[step] = \
+                    min(tStepArr[step-1]*fireParams["fInc"],fireParams["dtMax"])
+                alphaArr[step] = alphaArr[step-1]*fireParams["fAlpha"]
             else:
-                tStepArr[step,ptIter] = \
-                    max(tStepArr[step-1,ptIter]*fireParams["fDecel"],fireParams["dtMin"])
-                alphaArr[step,ptIter] = fireParams["aStart"]
-                stepsSinceReset[ptIter] = 0
+                tStepArr[step] = tStepArr[step-1]
+                alphaArr[step] = alphaArr[step-1]
+        else:
+            alphaArr[step] = fireParams["aStart"]
+            if(step > fireParams["minDecelIter"]):
+                tStepArr[step] = \
+                    max(tStepArr[step-1]*fireParams["fDecel"],fireParams["dtMin"])
+            else:
+                tStepArr[step] = tStepArr[step-1]
+            self.allPts[step-1] = self.allPts[step-1] - 0.5*tStepArr[step]*self.allVelocities[step-1,:,:]
+            self.allVelocities[step-1] = 0.0
+            
+        #Semi-implicit Euler integration
+        self.allVelocities[step] = self.allVelocities[step-1] + tStepArr[step] * self.allForces[step-1]
+        
+        #For mixing
+        vdotv = 0.0
+        fdotf = 0.0
+        for ptIter in range(self.nPts):
+            vdotv += np.dot(self.allVelocities[step,ptIter],self.allVelocities[step,ptIter])
+            fdotf += np.dot(self.allForces[step-1,ptIter],self.allForces[step-1,ptIter])
+        #Only vanishes if net force on all particles is zero, in which case this is zeroed out
+        #later anyways. Handled here to prevent nuisance exception-throwing
+        if fdotf > 10**(-16):
+            scale = np.sqrt(vdotv/fdotf)
+        else:
+            scale = 0.0
+        
+        self.allVelocities[step] = (1-alphaArr[step])*self.allVelocities[step] + \
+            alphaArr[step] * scale * self.allForces[step-1]
+        
+        shift = tStepArr[step]*self.allVelocities[step]
+        
+        for ptIter in range(self.nPts):
+            for dimIter in range(self.nDims):
+                if(abs(shift[ptIter,dimIter])>fireParams["maxmove"][dimIter]):
+                    shift[ptIter] = shift[ptIter] * \
+                        fireParams["maxmove"][dimIter]/abs(shift[ptIter,dimIter])
+        
+        self.allPts[step] = self.allPts[step-1] + shift
+        self.allForces[step] = self.nebObj.compute_force(self.allPts[step])
         
         return tStepArr, alphaArr, stepsSinceReset
     
     def _check_early_stop(self,currentIter,stopParams):
+        """
+        Computes standard deviation in location of every image over the previous
+        stopParams["nStabIters"] iterations. Returns True if std is less than
+        stopParams["stabPerc"] for every image.
+
+        Parameters
+        ----------
+        currentIter : TYPE
+            DESCRIPTION.
+        stopParams : TYPE
+            DESCRIPTION.
+
+        Returns
+        -------
+        ret : TYPE
+            DESCRIPTION.
+
+        """
         ret = False
         
         startCheckIter = stopParams["startCheckIter"]
@@ -1030,18 +1099,28 @@ class VerletMinimization:
         nStabIters = stopParams["nStabIters"]
         checkFreq = stopParams["checkFreq"]
         
-        #TODO: could cache energies/masses for use here
         if (currentIter >= startCheckIter) and (currentIter % checkFreq == 0):
-            tfs = np.array([self.nebObj.target_func(pt,self.nebObj.potential,self.nebObj.mass)[0] \
-                            for pt in self.allPts[currentIter-nStabIters:currentIter]])
-            mean = np.mean(tfs)
-            std = np.std(tfs)
-            
-            if std <= stabPerc:# * mean:
+            std = np.std(self.allPts[currentIter-nStabIters:currentIter],axis=0)
+            if np.all(std <= stabPerc):
                 ret = True
         
         return ret
     
+    def _check_early_abort(self,currentIter,breakParams):
+        
+        ret = False
+        
+        startCheckIter = breakParams["startCheckIter"]
+        allowedVariance = breakParams["variance"]
+        nStabIters = breakParams["nStabIters"]
+        checkFreq = breakParams["checkFreq"]
+        
+        if (currentIter >= startCheckIter) and (currentIter % checkFreq == 0):
+            std = np.std(self.allPts[currentIter-nStabIters:currentIter],axis=0)
+            if np.any(std >= allowedVariance):
+                ret = True
+        return ret
+    
 class EulerLagrangeSolver:
     """
     :Maintainer: Daniel
@@ -1540,7 +1619,6 @@ class DynamicProgramming:
     def __init__(self,initialPoint,coordMeshTuple,potArr,inertArr=None,allowedMask=None,\
                  target_func=TargetFunctions.action,allowedEndpoints=None,\
                  trimVals=[10**(-4),None],logLevel=1,fName=None,logFreq=50):
-        warnings.warn("DynamicProgramming class is still under development")
         self.initialPoint = initialPoint
         self.coordMeshTuple = coordMeshTuple
         self.uniqueCoords = [np.unique(c) for c in self.coordMeshTuple]
